# /usr/bin/env python3
# -*- coding:utf-8 -*-

"""
 For `python -m pigit`.
 When you run a python module directly, the interpreter
 looks for the `__main__.py` file in the module.
"""

from pigit import main

if __name__ == "__main__":
    main()
